/*
 * Copyright (C) 1996-1997 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)lxrun.c	1.8	99/02/25 SMI"

#include	<errno.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	"lxrun.h"
#include	"version.h"

char	emulator[PATH_MAX];	/* Contains full pathname of invoked lxrun */

extern void trace_init(void);
extern void lxrun_init(void);

extern char **environ;

static void
usage(const char *prog)
{
	fprintf(stderr, "Linux Emulator version %s, ", build_version);
	fprintf(stderr, "built on %s at %s\n", build_date, build_time);
	fprintf(stderr, "usage: %s [-0 argv0] linux_prog [args ...]\n", prog);
	exit(1);
}

/*
 * LINUX emulator for x86 UNIX
 */

int
main(int argc, char **argv, char **envp)
{
	char *pathenv = getenv("PATH");
	char *emul;			/* Path to the lxrun binary. */
	char *prog;			/* Path to the Linux program binary. */
	char *argv0 = NULL;		/* argv[0] to call program with. */
	int c;
	char	real_interp[PATH_MAX];
	int	ac = 0;
	char	**av;
#ifdef LXRUN_AUTO_PATH_BEHAVIOR
	char qualified[PATH_MAX];
#endif

	emul = argv[0];
	while ((c = getopt(argc, argv, "0:")) != EOF) {
		switch (c) {
		case '0':
			argv0 = optarg;
			break;
		default:
			usage(emul);
			break;
		}
	}
	argc -= optind; argv += optind;

	if (argc < 1)
		usage(emul);

	/*
	 * argv[0] is now the path to the Linux binary.  But if we were
	 * called with a different argv0 (e.g. by lx_execve), stick that
	 * into the argv[] array now.
	 */
	prog = argv[0];
	if (argv0 != NULL)
		argv[0] = argv0;

	find_emulator(pathenv, emulator, emul);

	trace_init();
	lxrun_init();

#ifdef LXRUN_AUTO_PATH_BEHAVIOR
	/*
	** If the filename we specify does not contain a slash,
	** check for it in our path.
	*/
	if(!strchr(prog, '/'))
		{
		/* We use a copy of pathenv since strtok is destructive. */
		char	*pathcopy = strdup(pathenv);
		char	*path = pathcopy;

		while((path = strtok(path, ":")))
			{
			strcpy(qualified, path);
			if(qualified[0]=='\0')
				strcpy(qualified, ".");
			strcat(qualified, "/");
			strcat(qualified, prog);

			/* Lxrun can't load the binary w/o read permissions. */
#if defined(SOLARIS)
			/*
			 * Don't want to have to pull in libgen.
			 */
			if(access(qualified, R_OK | 010) == 0 &&
			   access(qualified, X_OK | 010) == 0)
#else
			if(eaccess(qualified, R_OK) == 0 &&
			   eaccess(qualified, X_OK) == 0)
#endif
				{
				prog = qualified;
				break;
				}

			path = NULL;	/* reset path ptr for strtok() */
			}

		free(pathcopy);
		}
#endif
	run_program(prog, argv, environ);

	/*
	 * like exec() a successful call to run_program() will transfer
	 * control to the new program and never return - if we get
	 * here then run_program() has failed in some way.
	 * See if it's a shell script first.
	 */
	 
	if (is_shell_script(prog, real_interp, sizeof (real_interp))) {
		while (argv[ac++])
			continue;

		/*
		 * Feed the script to the interpreter.
		 */
		av = alloca((ac + 4) * sizeof(char *));
		av[0] = emulator;
		av[1] = "-0";
		av[2] = argv[0];
		av[3] = real_interp;
		av[4] = prog;
		memcpy(&av[5], &argv[1], (ac - 1) * sizeof (char *));

		prog = emulator;
		argv = av;
	}

	/*
	 * It may just be that we don't have read perms on the binary (common
	 * in OpenServer natives) and just try to exec it.
	 */
	execve(prog, argv, environ);

	fatal(errno, "%s load failed", *argv);

	return 1;
}
